"""Typedspark: column-wise type annotations for pyspark DataFrames."""

from typedspark._core.column import Column
from typedspark._core.column_meta import ColumnMeta
from typedspark._core.dataset import DataSet
from typedspark._core.datatypes import (
    ArrayType,
    DayTimeIntervalType,
    DecimalType,
    MapType,
    StructType,
)
from typedspark._core.literaltype import IntervalType
from typedspark._schema.schema import MetaSchema, Schema
from typedspark._transforms.structtype_column import structtype_column
from typedspark._transforms.transform_to_schema import transform_to_schema
from typedspark._utils.create_dataset import (
    create_empty_dataset,
    create_partially_filled_dataset,
    create_structtype_row,
)
from typedspark._utils.databases import Catalogs, Database, Databases
from typedspark._utils.load_table import create_schema, load_table
from typedspark._utils.register_schema_to_dataset import register_schema_to_dataset

__all__ = [
    "ArrayType",
    "Catalogs",
    "Column",
    "ColumnMeta",
    "Database",
    "Databases",
    "DataSet",
    "DayTimeIntervalType",
    "DecimalType",
    "IntervalType",
    "MapType",
    "MetaSchema",
    "Schema",
    "StructType",
    "create_empty_dataset",
    "create_partially_filled_dataset",
    "create_structtype_row",
    "create_schema",
    "load_table",
    "register_schema_to_dataset",
    "structtype_column",
    "transform_to_schema",
]



"""Module containing classes and functions related to TypedSpark DataSets."""
from copy import deepcopy
from typing import (
    Any,
    Callable,
    Generic,
    List,
    Literal,
    Optional,
    Type,
    TypeVar,
    Union,
    get_args,
    overload,
)

from pyspark.sql import Column as SparkColumn
from pyspark.sql import DataFrame
from typing_extensions import Concatenate, ParamSpec

from typedspark._core.validate_schema import validate_schema
from typedspark._schema.schema import Schema

T = TypeVar("T", bound=Schema)
_ReturnType = TypeVar("_ReturnType", bound=DataFrame)  # pylint: disable=C0103
P = ParamSpec("P")


class DataSet(DataFrame, Generic[T]):
    """``DataSet`` subclasses pyspark ``DataFrame`` and hence has all the same
    functionality, with in addition the possibility to define a schema.

    .. code-block:: python

        class Person(Schema):
            name: Column[StringType]
            age: Column[LongType]

        def foo(df: DataSet[Person]) -> DataSet[Person]:
            # do stuff
            return df
    """

    def __new__(cls, dataframe: DataFrame) -> "DataSet[T]":
        """``__new__()`` instantiates the object (prior to ``__init__()``).

        Here, we simply take the provided ``df`` and cast it to a
        ``DataSet``. This allows us to bypass the ``DataFrame``
        constuctor in ``__init__()``, which requires parameters that may
        be difficult to access.
        """
        dataframe.__class__ = DataSet
        return dataframe  # type: ignore

    def __init__(self, dataframe: DataFrame):
        pass

    def __setattr__(self, name: str, value: Any) -> None:
        """Python base function that sets attributes.

        We listen here for the setting of ``__orig_class__``, which
        contains the ``Schema`` of the ``DataSet``. Note that this gets
        set after ``__new__()`` and ``__init__()`` are finished.
        """
        object.__setattr__(self, name, value)

        if name == "__orig_class__":
            orig_class_args = get_args(self.__orig_class__)
            if orig_class_args and issubclass(orig_class_args[0], Schema):
                self._schema_annotations: Type[Schema] = orig_class_args[0]
                validate_schema(
                    self._schema_annotations.get_structtype(),
                    deepcopy(self.schema),
                    self._schema_annotations.get_schema_name(),
                )
                self._add_schema_metadata()

    def _add_schema_metadata(self) -> None:
        """Adds the ``ColumnMeta`` comments as metadata to the ``DataSet``.

        Previously set metadata is deleted. Hence, if ``foo(dataframe: DataSet[A]) -> DataSet[B]``,
        then ``DataSet[B]`` will not inherrit any metadata from ``DataSet[A]``.

        Assumes validate_schema() in __setattr__() has been run.
        """
        for field in self._schema_annotations.get_structtype().fields:
            self.schema[field.name].metadata = field.metadata

    @property
    def typedspark_schema(self) -> Type[T]:
        """Returns the ``Schema`` of the ``DataSet``."""
        return self._schema_annotations  # type: ignore

    """The following functions are equivalent to their parents in ``DataFrame``, but since they
    don't affect the ``Schema``, we can add type annotations here. We're omitting docstrings,
    such that the docstring from the parent will appear."""

    def distinct(self) -> "DataSet[T]":  # pylint: disable=C0116
        return DataSet[self._schema_annotations](super().distinct())  # type: ignore

    def filter(self, condition) -> "DataSet[T]":  # pylint: disable=C0116
        return DataSet[self._schema_annotations](super().filter(condition))  # type: ignore

    @overload
    def join(  # type: ignore
        self,
        other: DataFrame,
        on: Optional[  # pylint: disable=C0103
            Union[str, List[str], SparkColumn, List[SparkColumn]]
        ] = ...,
        how: None = ...,
    ) -> DataFrame:
        ...  # pragma: no cover

    @overload
    def join(
        self,
        other: DataFrame,
        on: Optional[  # pylint: disable=C0103
            Union[str, List[str], SparkColumn, List[SparkColumn]]
        ] = ...,
        how: Literal["semi"] = ...,
    ) -> "DataSet[T]":
        ...  # pragma: no cover

    @overload
    def join(
        self,
        other: DataFrame,
        on: Optional[  # pylint: disable=C0103
            Union[str, List[str], SparkColumn, List[SparkColumn]]
        ] = ...,
        how: Optional[str] = ...,
    ) -> DataFrame:
        ...  # pragma: no cover

    def join(  # pylint: disable=C0116
        self,
        other: DataFrame,
        on: Optional[  # pylint: disable=C0103
            Union[str, List[str], SparkColumn, List[SparkColumn]]
        ] = None,
        how: Optional[str] = None,
    ) -> DataFrame:
        return super().join(other, on, how)  # type: ignore

    def orderBy(self, *args, **kwargs) -> "DataSet[T]":  # type: ignore  # noqa: N802, E501  # pylint: disable=C0116, C0103
        return DataSet[self._schema_annotations](super().orderBy(*args, **kwargs))  # type: ignore

    @overload
    def transform(
        self,
        func: Callable[Concatenate["DataSet[T]", P], _ReturnType],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> _ReturnType:
        ...  # pragma: no cover

    @overload
    def transform(self, func: Callable[..., "DataFrame"], *args: Any, **kwargs: Any) -> "DataFrame":
        ...  # pragma: no cover

    def transform(  # pylint: disable=C0116
        self, func: Callable[..., "DataFrame"], *args: Any, **kwargs: Any
    ) -> "DataFrame":
        return super().transform(func, *args, **kwargs)

    @overload
    def unionByName(  # noqa: N802  # pylint: disable=C0116, C0103
        self,
        other: "DataSet[T]",
        allowMissingColumns: Literal[False] = ...,  # noqa: N803
    ) -> "DataSet[T]":
        ...  # pragma: no cover

    @overload
    def unionByName(  # noqa: N802  # pylint: disable=C0116, C0103
        self,
        other: DataFrame,
        allowMissingColumns: bool = ...,  # noqa: N803
    ) -> DataFrame:
        ...  # pragma: no cover

    def unionByName(  # noqa: N802  # pylint: disable=C0116, C0103
        self,
        other: DataFrame,
        allowMissingColumns: bool = False,  # noqa: N803
    ) -> DataFrame:
        res = super().unionByName(other, allowMissingColumns)
        if isinstance(other, DataSet) and other._schema_annotations == self._schema_annotations:
            return DataSet[self._schema_annotations](res)  # type: ignore
        return res  # pragma: no cover



"""Module responsible for generating StructFields from Columns in a Schema."""
from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Annotated, Type, TypeVar, Union, get_args, get_origin

from pyspark.sql.types import ArrayType as SparkArrayType
from pyspark.sql.types import DataType
from pyspark.sql.types import DayTimeIntervalType as SparkDayTimeIntervalType
from pyspark.sql.types import DecimalType as SparkDecimalType
from pyspark.sql.types import MapType as SparkMapType
from pyspark.sql.types import StructField
from pyspark.sql.types import StructType as SparkStructType

from typedspark._core.column import Column
from typedspark._core.column_meta import ColumnMeta
from typedspark._core.datatypes import (
    ArrayType,
    DayTimeIntervalType,
    DecimalType,
    MapType,
    StructType,
    TypedSparkDataType,
)

if TYPE_CHECKING:  # pragma: no cover
    from typedspark._schema.schema import Schema

_DataType = TypeVar("_DataType", bound=DataType)  # pylint: disable=invalid-name


def get_structfield(
    name: str,
    column: Union[Type[Column[_DataType]], Annotated[Type[Column[_DataType]], ColumnMeta]],
) -> StructField:
    """Generates a ``StructField`` for a given ``Column`` in a ``Schema``."""
    meta = get_structfield_meta(column)

    return StructField(
        name=name,
        dataType=_get_structfield_dtype(column, name),
        nullable=True,
        metadata=meta.get_metadata(),
    )


def get_structfield_meta(
    column: Union[Type[Column[_DataType]], Annotated[Type[Column[_DataType]], ColumnMeta]]
) -> ColumnMeta:
    """Get the spark column metadata from the ``ColumnMeta`` data, when
    available."""
    return next((x for x in get_args(column) if isinstance(x, ColumnMeta)), ColumnMeta())


def _get_structfield_dtype(
    column: Union[Type[Column[_DataType]], Annotated[Type[Column[_DataType]], ColumnMeta]],
    colname: str,
) -> DataType:
    """Get the spark ``DataType`` from the ``Column`` type annotation."""
    origin = get_origin(column)
    if origin not in [Annotated, Column]:
        raise TypeError(f"Column {colname} needs to be of type Column or Annotated.")

    if origin == Annotated:
        column = _get_column_from_annotation(column, colname)

    args = get_args(column)
    dtype = _get_dtype(args[0], colname)
    return dtype


def _get_column_from_annotation(
    column: Annotated[Type[Column[_DataType]], ColumnMeta],
    colname: str,
) -> Type[Column[_DataType]]:
    """Takes an ``Annotation[Column[...], ...]`` and returns the
    ``Column[...]``."""
    column = get_args(column)[0]
    if get_origin(column) != Column:
        raise TypeError(f"Column {colname} needs to have a Column[] within Annotated[].")

    return column


def _get_dtype(dtype: Type[DataType], colname: str) -> DataType:
    """Takes a ``DataType`` class and returns a DataType object."""
    origin = get_origin(dtype)
    if origin == ArrayType:
        return _extract_arraytype(dtype, colname)
    if origin == MapType:
        return _extract_maptype(dtype, colname)
    if origin == StructType:
        return _extract_structtype(dtype)
    if origin == DecimalType:
        return _extract_decimaltype(dtype)
    if origin == DayTimeIntervalType:
        return _extract_daytimeintervaltype(dtype)
    if (
        inspect.isclass(dtype)
        and issubclass(dtype, DataType)
        and not issubclass(dtype, TypedSparkDataType)
    ):
        return dtype()

    raise TypeError(
        f"Column {colname} does not have a correctly formatted DataType as a parameter."
    )


def _extract_arraytype(arraytype: Type[DataType], colname: str) -> SparkArrayType:
    """Takes e.g. an ``ArrayType[StringType]`` and creates an
    ``ArrayType(StringType(), True)``."""
    params = get_args(arraytype)
    element_type = _get_dtype(params[0], colname)
    return SparkArrayType(element_type)


def _extract_maptype(maptype: Type[DataType], colname: str) -> SparkMapType:
    """Takes e.g. a ``MapType[StringType, StringType]`` and creates a ``
    MapType(StringType(), StringType(), True)``."""
    params = get_args(maptype)
    key_type = _get_dtype(params[0], colname)
    value_type = _get_dtype(params[1], colname)
    return SparkMapType(key_type, value_type)


def _extract_structtype(structtype: Type[DataType]) -> SparkStructType:
    """Takes a ``StructType[Schema]`` annotation and creates a
    ``StructType(schema_list)``, where ``schema_list`` contains all
    ``StructField()`` defined in the ``Schema``."""
    params = get_args(structtype)
    schema: Type[Schema] = params[0]
    return schema.get_structtype()


def _extract_decimaltype(decimaltype: Type[DataType]) -> SparkDecimalType:
    """Takes e.g. a ``DecimalType[Literal[10], Literal[12]]`` and returns
    ``DecimalType(10, 12)``."""
    params = get_args(decimaltype)
    key_type: int = _unpack_literal(params[0])
    value_type: int = _unpack_literal(params[1])
    return SparkDecimalType(key_type, value_type)


def _extract_daytimeintervaltype(daytimeintervaltype: Type[DataType]) -> SparkDayTimeIntervalType:
    """Takes e.g. a ``DayTimeIntervalType[Literal[1], Literal[2]]`` and returns
    ``DayTimeIntervalType(1, 2)``."""
    params = get_args(daytimeintervaltype)
    start_field: int = _unpack_literal(params[0])
    end_field: int = _unpack_literal(params[1])
    return SparkDayTimeIntervalType(start_field, end_field)


def _unpack_literal(literal):
    """Takes as input e.g. ``Literal[10]`` and returns ``10``."""
    return get_args(literal)[0]



"""Module containing classes and functions related to TypedSpark Schemas."""
import inspect
import re
from typing import Any, Dict, List, Optional, Type, Union, get_args, get_type_hints

from pyspark.sql import DataFrame
from pyspark.sql.types import DataType, StructType

from typedspark._core.column import Column
from typedspark._schema.dlt_kwargs import DltKwargs
from typedspark._schema.get_schema_definition import get_schema_definition_as_string
from typedspark._schema.structfield import get_structfield


class MetaSchema(type):
    """``MetaSchema`` is the metaclass of ``Schema``.

    It basically implements all functionality of ``Schema``. But since
    classes are typically considered more convenient than metaclasses,
    we provide ``Schema`` as the public interface.

    .. code-block:: python

        class A(Schema):
            a: Column[IntegerType]
            b: Column[StringType]

        DataSet[A](df)

    The class methods of ``Schema`` are described here.
    """

    _parent: Optional[Union[DataFrame, Column]] = None
    _current_id: Optional[int] = None
    _original_name: Optional[str] = None

    def __new__(cls, name: str, bases: Any, dct: Dict[str, Any]):
        cls._attributes = dir(cls)

        # initializes all uninitialied variables with a type annotation as None
        # this allows for auto-complete in Databricks notebooks (uninitialized variables
        # don't show up in auto-complete there).
        if "__annotations__" in dct.keys():
            extra = {attr: None for attr in dct["__annotations__"] if attr not in dct}
            dct = dict(dct, **extra)

        return type.__new__(cls, name, bases, dct)

    def __repr__(cls) -> str:
        return f"\n{str(cls)}"

    def __str__(cls) -> str:
        return cls.get_schema_definition_as_string(add_subschemas=False)

    def __getattribute__(cls, name: str) -> Any:
        """Python base function that gets attributes.

        We listen here for anyone getting a ``Column`` from the ``Schema``.
        Even though they're not explicitely instantiated, we can instantiate
        them here whenever someone attempts to get them. This allows us to do the following:

        .. code-block:: python

            class A(Schema):
                a: Column[IntegerType]

            (
                df.withColumn(A.a.str, lit(1))
                .select(A.a)
            )
        """
        if name.startswith("__") or name == "_attributes" or name in cls._attributes:
            return object.__getattribute__(cls, name)

        if name in get_type_hints(cls):
            return Column(
                name,
                dtype=cls._get_dtype(name),  # type: ignore
                parent=cls._parent,
                curid=cls._current_id,
            )

        raise TypeError(f"Schema {cls.get_schema_name()} does not have attribute {name}.")

    def _get_dtype(cls, name: str) -> Type[DataType]:
        """Returns the datatype of a column, e.g. Column[IntegerType] -> IntegerType."""
        column = get_type_hints(cls)[name]
        args = get_args(column)

        if not args:
            raise TypeError(
                f"Column {cls.get_schema_name()}.{name} does not have an annotated type."
            )

        dtype = args[0]
        return dtype

    def all_column_names(cls) -> List[str]:
        """Returns all column names for a given schema."""
        return list(get_type_hints(cls).keys())

    def all_column_names_except_for(cls, except_for: List[str]) -> List[str]:
        """Returns all column names for a given schema except for the columns
        specified in the ``except_for`` parameter."""
        return list(name for name in get_type_hints(cls).keys() if name not in except_for)

    def get_snake_case(cls) -> str:
        """Return the class name transformed into snakecase."""
        word = cls.get_schema_name()
        word = re.sub(r"([A-Z]+)([A-Z][a-z])", r"\1_\2", word)
        word = re.sub(r"([a-z\d])([A-Z])", r"\1_\2", word)
        word = word.replace("-", "_")
        return word.lower()

    def get_schema_definition_as_string(
        cls,
        schema_name: Optional[str] = None,
        include_documentation: bool = False,
        generate_imports: bool = True,
        add_subschemas: bool = True,
    ) -> str:
        """Return the code for the ``Schema`` as a string."""
        if schema_name is None:
            schema_name = cls.get_schema_name()
        return get_schema_definition_as_string(
            cls,  # type: ignore
            include_documentation,
            generate_imports,
            add_subschemas,
            schema_name,
        )

    def print_schema(
        cls,
        schema_name: Optional[str] = None,
        include_documentation: bool = False,
        generate_imports: bool = True,
        add_subschemas: bool = False,
    ):  # pragma: no cover
        """Print the code for the ``Schema``."""
        print(
            cls.get_schema_definition_as_string(
                schema_name=schema_name,
                include_documentation=include_documentation,
                generate_imports=generate_imports,
                add_subschemas=add_subschemas,
            )
        )

    def get_docstring(cls) -> Union[str, None]:
        """Returns the docstring of the schema."""
        return inspect.getdoc(cls)

    def get_structtype(cls) -> StructType:
        """Creates the spark StructType for the schema."""
        return StructType(
            [
                get_structfield(name, column)
                for name, column in get_type_hints(cls, include_extras=True).items()
            ]
        )

    def get_dlt_kwargs(cls, name: Optional[str] = None) -> DltKwargs:
        """Creates a representation of the ``Schema`` to be used by Delta Live
        Tables.

        .. code-block:: python

            @dlt.table(**DimPatient.get_dlt_kwargs())
            def table_definition() -> DataSet[DimPatient]:
                <your table definition here>
        """
        return {
            "name": name if name else cls.get_snake_case(),
            "comment": cls.get_docstring(),
            "schema": cls.get_structtype(),
        }

    def get_schema_name(cls):
        """Returns the name with which the schema was initialized."""
        return cls._original_name if cls._original_name else cls.__name__


class Schema(metaclass=MetaSchema):
    # pylint: disable=missing-class-docstring
    # Since docstrings are inherrited, and since we use docstrings to
    # annotate tables (see MetaSchema.get_dlt_kwargs()), we have chosen
    # to not add a docstring to the Schema class (otherwise the Schema
    # docstring would be added to any schema without a docstring).
    pass



"""Functionality for dealing with StructType columns."""

from typing import Dict, Optional, Type

from pyspark.sql import Column as SparkColumn
from pyspark.sql.functions import struct

from typedspark._core.column import Column
from typedspark._schema.schema import Schema
from typedspark._transforms.utils import add_nulls_for_unspecified_columns, convert_keys_to_strings


def structtype_column(
    schema: Type[Schema],
    transformations: Optional[Dict[Column, SparkColumn]] = None,
    fill_unspecified_columns_with_nulls: bool = False,
) -> SparkColumn:
    """Helps with creating new ``StructType`` columns of a certain schema, for
    example:

    .. code-block:: python

        transform_to_schema(
            df,
            Output,
            {
                Output.values: structtype_column(
                    Value,
                    {
                        Value.a: Input.a + 2,
                        ...
                    }
                )
            }
        )
    """
    _transformations = convert_keys_to_strings(transformations)

    if fill_unspecified_columns_with_nulls:
        _transformations = add_nulls_for_unspecified_columns(_transformations, schema)

    _transformations = _order_columns(_transformations, schema)

    return struct([v.alias(k) for k, v in _transformations.items()])


def _order_columns(
    transformations: Dict[str, SparkColumn], schema: Type[Schema]
) -> Dict[str, SparkColumn]:
    """Chispa's DataFrame comparer doesn't deal nicely with StructTypes whose columns
    are ordered differently, hence we order them the same as in the schema here."""
    transformations_ordered = {}
    for field in schema.get_structtype().fields:
        transformations_ordered[field.name] = transformations[field.name]

    return transformations_ordered



"""Loads all catalogs, databases and tables in a SparkSession."""

from abc import ABC
from datetime import datetime
from typing import Any, Optional, Tuple, TypeVar
from warnings import warn

from pyspark.sql import Row, SparkSession

from typedspark._core.dataset import DataSet
from typedspark._schema.schema import Schema
from typedspark._utils.camelcase import to_camel_case
from typedspark._utils.load_table import load_table

T = TypeVar("T", bound=Schema)


class Timeout(ABC):
    """Warns the user if loading databases or catalogs is taking too long."""

    _TIMEOUT_WARNING: str

    def __init__(self, silent: bool, n: int):  # pylint: disable=invalid-name
        self._start = datetime.now()
        self._silent = silent
        self._n = n

    def check_for_warning(self, i: int):  # pragma: no cover
        """Checks if a warning should be issued."""
        if self._silent:
            return

        diff = datetime.now() - self._start
        if diff.seconds > 10:
            warn(self._TIMEOUT_WARNING.format(i, self._n))
            self._silent = True


class DatabasesTimeout(Timeout):
    """Warns the user if Databases() is taking too long."""

    _TIMEOUT_WARNING = """
Databases() is taking longer than 10 seconds. So far, {} out of {} databases have been loaded.
If this is too slow, consider loading a single database using:

from typedspark import Database

db = Database(spark, db_name=...)
"""


class CatalogsTimeout(Timeout):
    """Warns the user if Catalogs() is taking too long."""

    _TIMEOUT_WARNING = """
Catalogs() is taking longer than 10 seconds. So far, {} out of {} catalogs have been loaded.
If this is too slow, consider loading a single catalog using:

from typedspark import Databases

db = Databases(spark, catalog_name=...)
"""


class Table:
    """Loads a table in a database."""

    def __init__(self, spark: SparkSession, db_name: str, table_name: str, is_temporary: bool):
        self._spark = spark
        self._db_name = db_name
        self._table_name = table_name
        self._is_temporary = is_temporary

    @property
    def str(self) -> str:
        """Returns the path to the table, e.g. ``default.person``.

        While temporary tables are always stored in the ``default`` db, they are saved and
        loaded directly from their table name, e.g. ``person``.

        Non-temporary tables are saved and loaded from their full name, e.g.
        ``default.person``.
        """
        if self._is_temporary:
            return self._table_name

        return f"{self._db_name}.{self._table_name}"

    def load(self) -> Tuple[DataSet[T], T]:
        """Loads the table as a DataSet[T] and returns the schema."""
        return load_table(  # type: ignore
            self._spark,
            self.str,
            to_camel_case(self._table_name),
        )

    def __call__(self, *args: Any, **kwds: Any) -> Tuple[DataSet[T], T]:
        return self.load()


class Database:
    """Loads all tables in a database."""

    def __init__(self, spark: SparkSession, db_name: str, catalog_name: Optional[str] = None):
        if catalog_name is None:
            self._db_name = db_name
        else:
            self._db_name = f"{catalog_name}.{db_name}"

        tables = spark.sql(f"show tables from {self._db_name}").collect()
        for table in tables:
            table_name = table.tableName
            self.__setattr__(
                table_name,
                Table(spark, self._db_name, table_name, table.isTemporary),
            )

    @property
    def str(self) -> str:
        """Returns the database name."""
        return self._db_name


class Databases:
    """Loads all databases and tables in a SparkSession."""

    def __init__(
        self, spark: SparkSession, silent: bool = False, catalog_name: Optional[str] = None
    ):
        if catalog_name is None:
            query = "show databases"
        else:
            query = f"show databases in {catalog_name}"

        databases = spark.sql(query).collect()
        timeout = DatabasesTimeout(silent, n=len(databases))

        for i, database in enumerate(databases):
            timeout.check_for_warning(i)
            db_name = self._extract_db_name(database)
            self.__setattr__(db_name, Database(spark, db_name, catalog_name))

    def _extract_db_name(self, database: Row) -> str:
        """Extracts the database name from a Row.

        Old versions of Spark use ``databaseName``, newer versions use ``namespace``.
        """
        if hasattr(database, "databaseName"):  # pragma: no cover
            return database.databaseName
        if hasattr(database, "namespace"):
            return database.namespace

        raise ValueError(f"Could not find database name in {database}.")  # pragma: no cover


class Catalogs:
    """Loads all catalogs, databases and tables in a SparkSession."""

    def __init__(self, spark: SparkSession, silent: bool = False):
        catalogs = spark.sql("show catalogs").collect()
        timeout = CatalogsTimeout(silent, n=len(catalogs))

        for i, catalog in enumerate(catalogs):
            timeout.check_for_warning(i)
            catalog_name: str = catalog.catalog
            self.__setattr__(
                catalog_name,
                Databases(spark, silent=True, catalog_name=catalog_name),
            )



"""A representation of the ``Schema`` to be used by Delta Live Tables."""

from typing import Optional, TypedDict

from pyspark.sql.types import StructType


class DltKwargs(TypedDict):
    """A representation of the ``Schema`` to be used by Delta Live Tables.

    .. code-block:: python

        @dlt.table(**Person.get_dlt_kwargs())
        def table_definition() -> DataSet[Person]:
            <your table definition here>
    """

    name: str
    comment: Optional[str]
    schema: StructType



"""Builds an import statement for everything imported by a given ``Schema``."""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Type, get_args, get_origin, get_type_hints

from pyspark.sql.types import DataType

from typedspark._core.datatypes import (
    ArrayType,
    DayTimeIntervalType,
    DecimalType,
    MapType,
    StructType,
    TypedSparkDataType,
)

if TYPE_CHECKING:  # pragma: no cover
    from typedspark._schema.schema import Schema


def get_schema_imports(schema: Type[Schema], include_documentation: bool) -> str:
    """Builds an import statement for everything imported by the ``Schema``."""
    dtypes = _get_imported_dtypes(schema)
    return _build_import_string(dtypes, include_documentation)


def _get_imported_dtypes(schema: Type[Schema]) -> set[Type[DataType]]:
    """Returns a set of DataTypes that are imported by the given schema."""
    encountered_datatypes: set[Type[DataType]] = set()
    for column in get_type_hints(schema).values():
        args = get_args(column)
        if not args:
            continue

        dtype = args[0]
        encountered_datatypes |= _process_datatype(dtype)

    return encountered_datatypes


def _process_datatype(dtype: Type[DataType]) -> set[Type[DataType]]:
    """Returns a set of DataTypes that are imported for a given DataType.

    Handles nested DataTypes recursively.
    """
    encountered_datatypes: set[Type[DataType]] = set()

    origin: Optional[Type[DataType]] = get_origin(dtype)
    if origin:
        encountered_datatypes.add(origin)
    else:
        encountered_datatypes.add(dtype)

    if origin == MapType:
        key, value = get_args(dtype)
        encountered_datatypes |= _process_datatype(key)
        encountered_datatypes |= _process_datatype(value)

    if origin == ArrayType:
        element = get_args(dtype)[0]
        encountered_datatypes |= _process_datatype(element)

    if get_origin(dtype) == StructType:
        subschema = get_args(dtype)[0]
        encountered_datatypes |= _get_imported_dtypes(subschema)

    return encountered_datatypes


def _build_import_string(
    encountered_datatypes: set[Type[DataType]], include_documentation: bool
) -> str:
    """Returns a multiline string with the imports required for the given
    encountered_datatypes.

    Import sorting is applied.

    If the schema uses IntegerType, BooleanType, StringType, this functions result would be

    .. code-block:: python

        from pyspark.sql.types import BooleanType, IntegerType, StringType

        from typedspark import Column, Schema
    """
    return (
        _typing_imports(encountered_datatypes, include_documentation)
        + _pyspark_imports(encountered_datatypes)
        + _typedspark_imports(encountered_datatypes, include_documentation)
    )


def _typing_imports(encountered_datatypes: set[Type[DataType]], include_documentation: bool) -> str:
    """Returns the import statement for the typing library."""
    imports = []

    if any([dtype == DecimalType for dtype in encountered_datatypes]):
        imports += ["Literal"]

    if include_documentation:
        imports += ["Annotated"]

    if len(imports) > 0:
        imports = sorted(imports)
        imports_string = ", ".join(imports)  # type: ignore
        return f"from typing import {imports_string}\n\n"

    return ""


def _pyspark_imports(encountered_datatypes: set[Type[DataType]]) -> str:
    """Returns the import statement for the pyspark library."""
    dtypes = sorted(
        [
            dtype.__name__
            for dtype in encountered_datatypes
            if not issubclass(dtype, TypedSparkDataType)
        ]
    )

    if len(dtypes) > 0:
        dtypes_string = ", ".join(dtypes)
        return f"from pyspark.sql.types import {dtypes_string}\n\n"

    return ""


def _typedspark_imports(
    encountered_datatypes: set[Type[DataType]], include_documentation: bool
) -> str:
    """Returns the import statement for the typedspark library."""
    dtypes = [
        dtype.__name__ for dtype in encountered_datatypes if issubclass(dtype, TypedSparkDataType)
    ] + ["Column", "Schema"]

    if any([dtype == DayTimeIntervalType for dtype in encountered_datatypes]):
        dtypes += ["IntervalType"]

    if include_documentation:
        dtypes.append("ColumnMeta")

    dtypes = sorted(dtypes)

    dtypes_string = ", ".join(dtypes)
    return f"from typedspark import {dtypes_string}\n\n\n"



"""Functions for loading `DataSet` and `Schema` in notebooks."""

import re
from typing import Dict, Literal, Optional, Tuple, Type

from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.types import ArrayType as SparkArrayType
from pyspark.sql.types import DataType
from pyspark.sql.types import DayTimeIntervalType as SparkDayTimeIntervalType
from pyspark.sql.types import DecimalType as SparkDecimalType
from pyspark.sql.types import MapType as SparkMapType
from pyspark.sql.types import StructType as SparkStructType

from typedspark._core.column import Column
from typedspark._core.dataset import DataSet
from typedspark._core.datatypes import (
    ArrayType,
    DayTimeIntervalType,
    DecimalType,
    MapType,
    StructType,
)
from typedspark._schema.schema import MetaSchema, Schema
from typedspark._utils.camelcase import to_camel_case
from typedspark._utils.register_schema_to_dataset import register_schema_to_dataset


def _replace_illegal_column_names(dataframe: DataFrame) -> DataFrame:
    """Replaces illegal column names with a legal version."""
    mapping = _create_mapping(dataframe)

    for column, column_renamed in mapping.items():
        if column != column_renamed:
            dataframe = dataframe.withColumnRenamed(column, column_renamed)

    return dataframe


def _create_mapping(dataframe: DataFrame) -> Dict[str, str]:
    """Checks if there are duplicate columns after replacing illegal characters."""
    mapping = {column: _replace_illegal_characters(column) for column in dataframe.columns}
    renamed_columns = list(mapping.values())
    duplicates = {
        column: column_renamed
        for column, column_renamed in mapping.items()
        if renamed_columns.count(column_renamed) > 1
    }

    if len(duplicates) > 0:
        raise ValueError(
            "You're trying to dynamically generate a Schema from a DataFrame. "
            + "However, typedspark has detected that the DataFrame contains duplicate columns "
            + "after replacing illegal characters (e.g. whitespaces, dots, etc.).\n"
            + "The folowing columns have lead to duplicates:\n"
            + f"{duplicates}\n\n"
            + "Please rename these columns in your DataFrame."
        )

    return mapping


def _replace_illegal_characters(column_name: str) -> str:
    """Replaces illegal characters in a column name with an underscore."""
    return re.sub("[^A-Za-z0-9]", "_", column_name)


def _create_schema(structtype: SparkStructType, schema_name: Optional[str] = None) -> Type[Schema]:
    """Dynamically builds a ``Schema`` based on a ``DataFrame``'s
    ``StructType``"""
    type_annotations = {}
    attributes: Dict[str, None] = {}
    for column in structtype:
        name = column.name
        data_type = _extract_data_type(column.dataType, name)
        type_annotations[name] = Column[data_type]  # type: ignore
        attributes[name] = None

    if not schema_name:
        schema_name = "DynamicallyLoadedSchema"

    schema = MetaSchema(schema_name, tuple([Schema]), attributes)
    schema.__annotations__ = type_annotations

    return schema  # type: ignore


def _extract_data_type(dtype: DataType, name: str) -> Type[DataType]:
    """Given an instance of a ``DataType``, it extracts the corresponding
    ``DataType`` class, potentially including annotations (e.g.
    ``ArrayType[StringType]``)."""
    if isinstance(dtype, SparkArrayType):
        element_type = _extract_data_type(dtype.elementType, name)
        return ArrayType[element_type]  # type: ignore

    if isinstance(dtype, SparkMapType):
        key_type = _extract_data_type(dtype.keyType, name)
        value_type = _extract_data_type(dtype.valueType, name)
        return MapType[key_type, value_type]  # type: ignore

    if isinstance(dtype, SparkStructType):
        subschema = _create_schema(dtype, to_camel_case(name))
        return StructType[subschema]  # type: ignore

    if isinstance(dtype, SparkDayTimeIntervalType):
        start_field = dtype.startField
        end_field = dtype.endField
        return DayTimeIntervalType[Literal[start_field], Literal[end_field]]  # type: ignore

    if isinstance(dtype, SparkDecimalType):
        precision = dtype.precision
        scale = dtype.scale
        return DecimalType[Literal[precision], Literal[scale]]  # type: ignore

    return type(dtype)


def create_schema(
    dataframe: DataFrame, schema_name: Optional[str] = None
) -> Tuple[DataSet[Schema], Type[Schema]]:
    """This function inferres a ``Schema`` in a notebook based on a the provided ``DataFrame``.

    This allows for autocompletion on column names, amongst other
    things.

    .. code-block:: python

        df, Person = create_schema(df)
    """
    dataframe = _replace_illegal_column_names(dataframe)
    schema = _create_schema(dataframe.schema, schema_name)
    dataset = DataSet[schema](dataframe)  # type: ignore
    schema = register_schema_to_dataset(dataset, schema)
    return dataset, schema


def load_table(
    spark: SparkSession, table_name: str, schema_name: Optional[str] = None
) -> Tuple[DataSet[Schema], Type[Schema]]:
    """This function loads a ``DataSet``, along with its inferred ``Schema``,
    in a notebook.

    This allows for autocompletion on column names, amongst other
    things.

    .. code-block:: python

        df, Person = load_table(spark, "path.to.table")
    """
    dataframe = spark.table(table_name)
    return create_schema(dataframe, schema_name)



"""Defines ``LiteralTypes``, e.g. ``IntervalType.DAY``, that map their class attribute to a
``Literal`` integer. Can be used for example in ``DayTimeIntervalType``."""
from typing import Dict, Literal


class LiteralType:
    """Base class for literal types, that map their class attribute to a Literal
    integer."""

    @classmethod
    def get_dict(cls) -> Dict[str, str]:
        """Returns a dictionary mapping e.g. "IntervalType.DAY" -> "Literal[0]"."""
        dictionary = {}
        for key, value in cls.__dict__.items():
            if key.startswith("_"):
                continue

            key = f"{cls.__name__}.{key}"
            value = str(value).replace("typing.", "")

            dictionary[key] = value

        return dictionary

    @classmethod
    def get_inverse_dict(cls) -> Dict[str, str]:
        """Returns a dictionary mapping e.g. "Literal[0]" -> "IntervalType.DAY"."""
        return {v: k for k, v in cls.get_dict().items()}


class IntervalType(LiteralType):
    """Interval types for ``DayTimeIntervalType``."""

    DAY = Literal[0]
    HOUR = Literal[1]
    MINUTE = Literal[2]
    SECOND = Literal[3]



"""Module containing functions that are related to transformations to DataSets."""
from functools import reduce
from typing import Dict, Optional, Type, TypeVar, Union

from pyspark.sql import Column as SparkColumn
from pyspark.sql import DataFrame

from typedspark._core.column import Column
from typedspark._core.dataset import DataSet
from typedspark._schema.schema import Schema
from typedspark._transforms.rename_duplicate_columns import RenameDuplicateColumns
from typedspark._transforms.utils import add_nulls_for_unspecified_columns, convert_keys_to_strings

T = TypeVar("T", bound=Schema)


def _do_transformations(dataframe: DataFrame, transformations: Dict[str, SparkColumn]) -> DataFrame:
    """Performs the transformations on the provided DataFrame."""
    return reduce(
        lambda acc, key: DataFrame.withColumn(acc, key, transformations[key]),
        transformations.keys(),
        dataframe,
    )


def _rename_temporary_keys_to_original_keys(
    dataframe: DataFrame, problematic_key_mapping: Dict[str, str]
) -> DataFrame:
    """Renames the temporary keys back to the original keys."""
    return reduce(
        lambda acc, key: DataFrame.withColumnRenamed(acc, problematic_key_mapping[key], key),
        problematic_key_mapping.keys(),
        dataframe,
    )


def transform_to_schema(
    dataframe: DataFrame,
    schema: Type[T],
    transformations: Optional[Dict[Column, SparkColumn]] = None,
    fill_unspecified_columns_with_nulls: bool = False,
) -> DataSet[T]:
    """On the provided DataFrame ``df``, it performs the ``transformations`` (if
    provided), and subsequently subsets the resulting DataFrame to the columns specified
    in ``schema``.

    .. code-block:: python

        transform_to_schema(
            df_a.join(df_b, A.a == B.f),
            AB,
            {
                AB.a: A.a + 3,
                AB.b: A.b + 7,
                AB.i: B.i - 5,
                AB.j: B.j + 1,
            }
        )
    """
    transform: Union[dict[str, SparkColumn], RenameDuplicateColumns]
    transform = convert_keys_to_strings(transformations)

    if fill_unspecified_columns_with_nulls:
        transform = add_nulls_for_unspecified_columns(transform, schema, dataframe.columns)

    transform = RenameDuplicateColumns(transform, schema, dataframe.columns)

    return DataSet[schema](  # type: ignore
        dataframe.transform(_do_transformations, transform.transformations)
        .drop(*transform.temporary_key_mapping.keys())
        .transform(_rename_temporary_keys_to_original_keys, transform.temporary_key_mapping)
        .select(*schema.all_column_names())
    )



"""Utility function for converting from snake case to camel case."""


def to_camel_case(name: str) -> str:
    """Converts a string to camel case."""
    return "".join([word.capitalize() for word in name.split("_")])



"""Module containing functions that are related to registering schema's to DataSets."""
import itertools
from typing import Type, TypeVar

from typedspark._core.dataset import DataSet
from typedspark._schema.schema import Schema

T = TypeVar("T", bound=Schema)


def _counter(count: itertools.count = itertools.count()):
    return next(count)


def register_schema_to_dataset(dataframe: DataSet[T], schema: Type[T]) -> Type[T]:
    """Helps combat column ambiguity. For example:

    .. code-block:: python

        class Person(Schema):
            id: Column[IntegerType]
            name: Column[StringType]

        class Job(Schema):
            id: Column[IntegerType]
            salary: Column[IntegerType]

        class PersonWithJob(Person, Job):
            pass

        def foo(df_a: DataSet[Person], df_b: DataSet[Job]) -> DataSet[PersonWithJob]:
            return DataSet[PersonWithSalary](
                df_a.join(
                    df_b,
                    Person.id == Job.id
                )
            )

    Calling ``foo()`` would result in a ``AnalysisException``, because Spark can't figure out
    whether ``id`` belongs to ``df_a`` or ``df_b``. To deal with this, you need to register
    your ``Schema`` to the ``DataSet``.

    .. code-block:: python

        from typedspark import register_schema_to_dataset

        def foo(df_a: DataSet[Person], df_b: DataSet[Job]) -> DataSet[PersonWithSalary]:
            person = register_schema_to_dataset(df_a, Person)
            job = register_schema_to_dataset(df_b, Job)
            return DataSet[PersonWithSalary](
                df_a.join(
                    df_b,
                    person.id == job.id
                )
            )
    """

    class LinkedSchema(schema):  # type: ignore  # pylint: disable=missing-class-docstring
        _parent = dataframe
        _current_id = _counter()
        _original_name = schema.get_schema_name()

    return LinkedSchema  # type: ignore



"""Metadata for ``Column`` objects that can be accessed during runtime."""
from dataclasses import dataclass
from typing import Dict, Optional


@dataclass
class ColumnMeta:
    """Contains the metadata for a ``Column``. Used as:

    .. code-block:: python

        class A(Schema):
            a: Annotated[
                Column[IntegerType],
                ColumnMeta(
                    comment="This is a comment",
                )
            ]
    """

    comment: Optional[str] = None

    def get_metadata(self) -> Optional[Dict[str, str]]:
        """Returns the metadata of this column."""
        return {"comment": self.comment} if self.comment else None



"""Here, we make our own definitions of ``MapType``, ``ArrayType`` and
``StructType`` in order to allow e.g. for ``ArrayType[StringType]``."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Generic, Type, TypeVar

from pyspark.sql.types import DataType

if TYPE_CHECKING:  # pragma: no cover
    from typedspark._schema.schema import Schema

    _Schema = TypeVar("_Schema", bound=Schema)
else:
    _Schema = TypeVar("_Schema")

_KeyType = TypeVar("_KeyType", bound=DataType)  # pylint: disable=invalid-name
_ValueType = TypeVar("_ValueType", bound=DataType)  # pylint: disable=invalid-name
_Precision = TypeVar("_Precision", bound=int)  # pylint: disable=invalid-name
_Scale = TypeVar("_Scale", bound=int)  # pylint: disable=invalid-name
_StartField = TypeVar("_StartField", bound=int)  # pylint: disable=invalid-name
_EndField = TypeVar("_EndField", bound=int)  # pylint: disable=invalid-name


class TypedSparkDataType(DataType):
    """Base class for typedspark specific ``DataTypes``."""

    @classmethod
    def get_name(cls) -> str:
        """Return the name of the type."""
        return cls.__name__


class StructTypeMeta(type):
    """Initializes the schema attribute as None.

    This allows for auto-complete in Databricks notebooks (uninitialized variables don't
    show up in auto-complete there).
    """

    def __new__(cls, name: str, bases: Any, dct: Dict[str, Any]):
        dct["schema"] = None
        return super().__new__(cls, name, bases, dct)


class StructType(Generic[_Schema], TypedSparkDataType, metaclass=StructTypeMeta):
    """Allows for type annotations such as:

    .. code-block:: python

        class Job(Schema):
            position: Column[StringType]
            salary: Column[LongType]

        class Person(Schema):
            job: Column[StructType[Job]]
    """

    schema: Type[_Schema]


class MapType(Generic[_KeyType, _ValueType], TypedSparkDataType):
    """Allows for type annotations such as.

    .. code-block:: python

        class Basket(Schema):
            items: Column[MapType[StringType, StringType]]
    """


class ArrayType(Generic[_ValueType], TypedSparkDataType):
    """Allows for type annotations such as.

    .. code-block:: python

        class Basket(Schema):
            items: Column[ArrayType[StringType]]
    """


class DecimalType(Generic[_Precision, _Scale], TypedSparkDataType):
    """Allows for type annotations such as.

    .. code-block:: python

        class Numbers(Schema):
            number: Column[DecimalType[Literal[10], Literal[0]]]
    """


class DayTimeIntervalType(Generic[_StartField, _EndField], TypedSparkDataType):
    """Allows for type annotations such as.

    .. code-block:: python

        class TimeInterval(Schema):
            interval: Column[DayTimeIntervalType[IntervalType.HOUR, IntervalType.SECOND]
    """



"""Module to output a string with the ``Schema`` definition of a given
``DataFrame``."""
from __future__ import annotations

import re
from typing import TYPE_CHECKING, Type, get_args, get_origin, get_type_hints

from typedspark._core.datatypes import DayTimeIntervalType, StructType, TypedSparkDataType
from typedspark._core.literaltype import IntervalType, LiteralType
from typedspark._schema.get_schema_imports import get_schema_imports

if TYPE_CHECKING:  # pragma: no cover
    from typedspark._schema.schema import Schema


def get_schema_definition_as_string(
    schema: Type[Schema],
    include_documentation: bool,
    generate_imports: bool,
    add_subschemas: bool,
    class_name: str = "MyNewSchema",
) -> str:
    """Return the code for a given ``Schema`` as a string.

    Typically used when you load a dataset using
    ``load_dataset_from_table()`` in a notebook and you want to save the
    schema in your code base. When ``generate_imports`` is True, the
    required imports for the schema are included in the string.
    """
    imports = get_schema_imports(schema, include_documentation) if generate_imports else ""
    schema_string = _build_schema_definition_string(
        schema, include_documentation, add_subschemas, class_name
    )

    return imports + schema_string


def _build_schema_definition_string(
    schema: Type[Schema],
    include_documentation: bool,
    add_subschemas: bool,
    class_name: str = "MyNewSchema",
) -> str:
    """Return the code for a given ``Schema`` as a string."""
    lines = f"class {class_name}(Schema):\n"

    if include_documentation:
        lines += _create_docstring(schema)

    lines += _add_lines_with_typehint(include_documentation, schema)

    if add_subschemas:
        lines += _add_subschemas(schema, add_subschemas, include_documentation)

    return lines


def _create_docstring(schema: Type[Schema]) -> str:
    """Create the docstring for a given ``Schema``."""
    if schema.get_docstring() is not None:
        docstring = f'    """{schema.get_docstring()}"""\n\n'
    else:
        docstring = '    """Add documentation here."""\n\n'
    return docstring


def _add_lines_with_typehint(include_documentation, schema):
    """Add a line with the typehint for each column in the ``Schema``."""
    lines = ""
    for col_name, col_type in get_type_hints(schema, include_extras=True).items():
        typehint, comment = _create_typehint_and_comment(col_type)

        if include_documentation:
            lines += f'    {col_name}: Annotated[{typehint}, ColumnMeta(comment="{comment}")]\n'
        else:
            lines += f"    {col_name}: {typehint}\n"
    return lines


def _create_typehint_and_comment(col_type) -> list[str]:
    """Create a typehint and comment for a given column."""
    typehint = (
        str(col_type)
        .replace("typedspark._core.column.", "")
        .replace("typedspark._core.datatypes.", "")
        .replace("typedspark._schema.schema.", "")
        .replace("pyspark.sql.types.", "")
        .replace("typing.", "")
    )
    typehint, comment = _extract_comment(typehint)
    typehint = _replace_literals(
        typehint, replace_literals_in=DayTimeIntervalType, replace_literals_by=IntervalType
    )
    return [typehint, comment]


def _extract_comment(typehint: str) -> tuple[str, str]:
    """Extract the comment from a typehint."""
    comment = ""
    if "Annotated" in typehint:
        match = re.search(r"Annotated\[(.*), '(.*)'\]", typehint)
        if match is not None:
            typehint, comment = match.groups()
    return typehint, comment


def _replace_literals(
    typehint: str,
    replace_literals_in: Type[TypedSparkDataType],
    replace_literals_by: Type[LiteralType],
) -> str:
    """Replace all Literals in a LiteralType, e.g.

    "DayTimeIntervalType[Literal[0], Literal[1]]" ->
    "DayTimeIntervalType[IntervalType.DAY, IntervalType.HOUR]"
    """
    mapping = replace_literals_by.get_inverse_dict()
    for original, replacement in mapping.items():
        typehint = _replace_literal(typehint, replace_literals_in, original, replacement)

    return typehint


def _replace_literal(
    typehint: str,
    replace_literals_in: Type[TypedSparkDataType],
    original: str,
    replacement: str,
) -> str:
    """Replaces a single Literal in a LiteralType, e.g.

    "DayTimeIntervalType[Literal[0], Literal[1]]" ->
    "DayTimeIntervalType[IntervalType.DAY, Literal[1]]"
    """
    return re.sub(
        rf"{replace_literals_in.get_name()}\[[^]]*\]",
        lambda x: x.group(0).replace(original, replacement),
        typehint,
    )


def _add_subschemas(schema: Type[Schema], add_subschemas: bool, include_documentation: bool) -> str:
    """Identifies whether any ``Column`` are of the ``StructType`` type and
    generates their schema recursively."""
    lines = ""
    for val in get_type_hints(schema).values():
        args = get_args(val)
        if not args:
            continue

        dtype = args[0]
        if get_origin(dtype) == StructType:
            lines += "\n\n"
            subschema: Type[Schema] = get_args(dtype)[0]
            lines += _build_schema_definition_string(
                subschema, include_documentation, add_subschemas, subschema.get_schema_name()
            )

    return lines



"""Util functions for typedspark._transforms."""
from typing import Dict, List, Optional, Type

from pyspark.sql import Column as SparkColumn
from pyspark.sql.functions import lit

from typedspark._core.column import Column
from typedspark._schema.schema import Schema


def add_nulls_for_unspecified_columns(
    transformations: Dict[str, SparkColumn],
    schema: Type[Schema],
    previously_existing_columns: Optional[List[str]] = None,
) -> Dict[str, SparkColumn]:
    """Takes the columns from the schema that are not present in the transformation
    dictionary and sets their values to Null (casted to the corresponding type defined
    in the schema)."""
    _previously_existing_columns = (
        [] if previously_existing_columns is None else previously_existing_columns
    )
    for field in schema.get_structtype().fields:
        if field.name not in transformations and field.name not in _previously_existing_columns:
            transformations[field.name] = lit(None).cast(field.dataType)

    return transformations


def convert_keys_to_strings(
    transformations: Optional[Dict[Column, SparkColumn]]
) -> Dict[str, SparkColumn]:
    """Takes the Column keys in transformations and converts them to strings."""
    if transformations is None:
        return {}

    _transformations = {k.str: v for k, v in transformations.items()}

    if len(transformations) != len(_transformations):
        raise ValueError(
            "The transformations dictionary requires columns with unique names as keys. "
            + "It is currently not possible to have ambiguous column names here, "
            + "even when used in combination with register_schema_to_dataset()."
        )

    return _transformations



"""Module that handles duplicate columns in the ``DataFrame``, that are also in the
schema (and hence in the resulting ``DataSet[Schema]``), but which are not handled by
the transformations dictionary."""

from typing import Dict, Final, Type
from uuid import uuid4

from pyspark.sql import Column as SparkColumn

from typedspark._schema.schema import Schema

ERROR_MSG: Final[
    str
] = """Columns {columns} are ambiguous.
Please specify the transformations for these columns explicitly, for example:

schema_a = register_schema_to_dataset(df_a, A)
schema_b = register_schema_to_dataset(df_b, B)

transform_to_schema(
    df_a.join(
        df_b,
        schema_a.id == schema_b.id
    ),
    C,
    {{
        C.id: schema_a.id,
    }}
)
"""


class RenameDuplicateColumns:
    """Class that handles duplicate columns in the DataFrame, that are also in the
    schema (and hence in the resulting ``DataSet[Schema]``), but which are not handled
    by the transformations dictionary.

    This class renames these duplicate columns to temporary names, such that we avoid
    ambiguous columns in the resulting ``DataSet[Schema]``.
    """

    def __init__(
        self,
        transformations: Dict[str, SparkColumn],
        schema: Type[Schema],
        dataframe_columns: list[str],
    ):
        self._temporary_key_mapping = self._create_temporary_key_mapping(
            transformations, dataframe_columns, schema
        )
        self._transformations = self._rename_keys_to_temporary_keys(transformations)

    def _create_temporary_key_mapping(
        self,
        transformations: Dict[str, SparkColumn],
        dataframe_columns: list[str],
        schema: Type[Schema],
    ) -> Dict[str, str]:
        """Creates a mapping for duplicate columns in the ``DataFrame`` to temporary
        names, such that we avoid ambiguous columns in the resulting
        ``DataSet[Schema]``."""
        duplicate_columns_in_dataframe = self._duplicates(dataframe_columns)
        schema_columns = set(schema.all_column_names())
        transformation_keys = set(transformations.keys())

        self._verify_that_all_duplicate_columns_will_be_handled(
            duplicate_columns_in_dataframe, schema_columns, transformation_keys
        )

        problematic_keys = duplicate_columns_in_dataframe & transformation_keys

        res = {}
        for key in problematic_keys:
            res[key] = self._find_temporary_name(key, dataframe_columns)

        return {k: f"my_temporary_typedspark_{k}" for k in problematic_keys}

    def _duplicates(self, lst: list) -> set:
        """Returns a set of the duplicates in the provided list."""
        return {x for x in lst if lst.count(x) > 1}

    def _verify_that_all_duplicate_columns_will_be_handled(
        self,
        duplicate_columns_in_dataframe: set[str],
        schema_columns: set[str],
        transformation_keys: set[str],
    ):
        """Raises an exception if there are duplicate columns in the ``DataFrame``, that
        are also in the schema (and hence in the resulting ``DataSet[Schema]``), but
        which are not handled by the ``transformation``s`` dictionary."""
        unhandled_columns = (duplicate_columns_in_dataframe & schema_columns) - transformation_keys
        if unhandled_columns:
            raise ValueError(ERROR_MSG.format(columns=unhandled_columns))

    def _find_temporary_name(self, colname: str, dataframe_columns: list[str]) -> str:
        """Appends a uuid to the column name to make sure the temporary name doesn't
        collide with any other column names."""
        name = colname
        num = 0
        while name in dataframe_columns:
            name = f"{colname}_with_temporary_uuid_{uuid4()}"
            num += 1
            if num > 100:
                raise Exception("Failed to find a temporary name.")  # pragma: no cover

        return name

    def _rename_keys_to_temporary_keys(
        self, transformations: Dict[str, SparkColumn]
    ) -> Dict[str, SparkColumn]:
        """Renames the keys in the transformations dictionary to temporary keys."""
        return {self._temporary_key_mapping.get(k, k): v for k, v in transformations.items()}

    @property
    def transformations(self) -> Dict[str, SparkColumn]:
        """Returns the transformations dictionary."""
        return self._transformations

    @property
    def temporary_key_mapping(self) -> Dict[str, str]:
        """Returns the temporary key mapping."""
        return self._temporary_key_mapping



"""Module containing functions related to creating a DataSet from scratch."""
from typing import Any, Dict, List, Type, TypeVar, Union, get_type_hints

from pyspark.sql import Row, SparkSession

from typedspark._core.column import Column
from typedspark._core.dataset import DataSet
from typedspark._schema.schema import Schema

T = TypeVar("T", bound=Schema)


def create_empty_dataset(spark: SparkSession, schema: Type[T], n_rows: int = 3) -> DataSet[T]:
    """Creates a ``DataSet`` with ``Schema`` schema, containing ``n_rows``
    rows, filled with ``None`` values.

    .. code-block:: python

        class Person(Schema):
            name: Column[StringType]
            age: Column[LongType]

        df = create_empty_dataset(spark, Person)
    """
    n_cols = len(get_type_hints(schema))
    rows = tuple([None] * n_cols)
    data = [rows] * n_rows
    spark_schema = schema.get_structtype()
    dataframe = spark.createDataFrame(data, spark_schema)
    return DataSet[schema](dataframe)  # type: ignore


def create_partially_filled_dataset(
    spark: SparkSession,
    schema: Type[T],
    data: Union[Dict[Column, List[Any]], List[Dict[Column, Any]]],
) -> DataSet[T]:
    """Creates a ``DataSet`` with ``Schema`` schema, where ``data`` can
    be defined in either of the following two ways:

    .. code-block:: python

        class Person(Schema):
            name: Column[StringType]
            age: Column[LongType]
            job: Column[StringType]

        df = create_partially_filled_dataset(
            spark,
            Person,
            {
                Person.name: ["John", "Jack", "Jane"],
                Person.age: [30, 40, 50],
            }
        )

    Or:

    .. code-block:: python

        df = create_partially_filled_dataset(
            spark,
            Person,
            [
                {Person.name: "John", Person.age: 30},
                {Person.name: "Jack", Person.age: 40},
                {Person.name: "Jane", Person.age: 50},
            ]
        )

    Any columns in the schema that are not present in the data will be
    initialized with ``None`` values.
    """
    if isinstance(data, list):
        col_data = _create_column_wise_data_from_list(schema, data)
    elif isinstance(data, dict):
        col_data = _create_column_wise_data_from_dict(schema, data)
    else:
        raise ValueError("The provided data is not a list or a dict.")

    row_data = zip(*col_data)
    spark_schema = schema.get_structtype()
    dataframe = spark.createDataFrame(row_data, spark_schema)
    return DataSet[schema](dataframe)  # type: ignore


def create_structtype_row(schema: Type[T], data: Dict[Column, Any]) -> Row:
    """Creates a ``Row`` with ``StructType`` schema, where ``data`` is a
    mapping from column to data in the respective column."""
    data_with_string_index = {k.str: v for k, v in data.items()}
    data_converted = {
        k: data_with_string_index[k] if k in data_with_string_index else None
        for k in get_type_hints(schema).keys()
    }
    return Row(**data_converted)


def _create_column_wise_data_from_dict(
    schema: Type[T], data: Dict[Column, List[Any]]
) -> List[List[Any]]:
    """Converts a dict of column to data to a list of lists, where each inner list
    contains the data for a column."""
    data_converted = {k.str: v for k, v in data.items()}
    n_rows_unique = {len(v) for _, v in data.items()}
    if len(n_rows_unique) > 1:
        raise ValueError("The number of rows in the provided data differs per column.")

    n_rows = list(n_rows_unique)[0]
    col_data = []
    for col in get_type_hints(schema).keys():
        if col in data_converted:
            col_data += [data_converted[col]]
        else:
            col_data += [[None] * n_rows]

    return col_data


def _create_column_wise_data_from_list(
    schema: Type[T], data: List[Dict[Column, Any]]
) -> List[List[Any]]:
    """Converts a list of dicts of column to data to a list of lists, where each inner
    list contains the data for a column."""
    data_converted = [{k.str: v for k, v in row.items()} for row in data]

    col_data = []
    for col in get_type_hints(schema).keys():
        col_data += [[row[col] if col in row else None for row in data_converted]]

    return col_data



"""Module containing functions that are related to validating schema's at runtime."""
from typing import Dict, Set

from pyspark.sql.types import ArrayType, DataType, MapType, StructField, StructType


def validate_schema(
    structtype_expected: StructType, structtype_observed: StructType, schema_name: str
) -> None:
    """Checks whether the expected and the observed StructType match."""
    expected = unpack_schema(structtype_expected)
    observed = unpack_schema(structtype_observed)

    check_names(set(expected.keys()), set(observed.keys()), schema_name)
    check_dtypes(expected, observed, schema_name)


def unpack_schema(schema: StructType) -> Dict[str, StructField]:
    """Converts the observed schema to a dictionary mapping column name to StructField.

    We ignore columns that start with ``__``.
    """
    res = {}
    for field in schema.fields:
        if field.name.startswith("__"):
            continue
        field.nullable = True
        field.metadata = {}
        res[field.name] = field

    return res


def check_names(names_expected: Set[str], names_observed: Set[str], schema_name: str) -> None:
    """Checks whether the observed and expected list of column names overlap.

    Is order insensitive.
    """
    diff = names_observed - names_expected
    if diff:
        raise TypeError(
            f"Data contains the following columns not present in schema {schema_name}: {diff}"
        )

    diff = names_expected - names_observed
    if diff:
        raise TypeError(
            f"Schema {schema_name} contains the following columns not present in data: {diff}"
        )


def check_dtypes(
    schema_expected: Dict[str, StructField],
    schema_observed: Dict[str, StructField],
    schema_name: str,
) -> None:
    """Checks for each column whether the observed and expected data type match.

    Is order insensitive.
    """
    for name, structfield_expected in schema_expected.items():
        structfield_observed = schema_observed[name]
        check_dtype(
            name,
            structfield_expected.dataType,
            structfield_observed.dataType,
            schema_name,
        )


def check_dtype(
    colname: str, dtype_expected: DataType, dtype_observed: DataType, schema_name: str
) -> None:
    """Checks whether the observed and expected data type match."""
    if dtype_expected == dtype_observed:
        return None

    if isinstance(dtype_expected, ArrayType) and isinstance(dtype_observed, ArrayType):
        return check_dtype(
            f"{colname}.element_type",
            dtype_expected.elementType,
            dtype_observed.elementType,
            schema_name,
        )

    if isinstance(dtype_expected, MapType) and isinstance(dtype_observed, MapType):
        check_dtype(
            f"{colname}.key",
            dtype_expected.keyType,
            dtype_observed.keyType,
            schema_name,
        )
        return check_dtype(
            f"{colname}.value",
            dtype_expected.valueType,
            dtype_observed.valueType,
            schema_name,
        )

    if isinstance(dtype_expected, StructType) and isinstance(dtype_observed, StructType):
        return validate_schema(dtype_expected, dtype_observed, f"{schema_name}.{colname}")

    raise TypeError(
        f"Column {colname} is of type {dtype_observed}, but {schema_name}.{colname} "
        + f"suggests {dtype_expected}."
    )



from setuptools import find_packages, setup


def get_requirements():
    with open("requirements.txt") as f:
        return f.read().splitlines()


def get_long_description():
    with open("README.rst", encoding="utf-8") as f:
        return f.read()


setup(
    name="typedspark",
    url="https://github.com/kaiko-ai/typedspark",
    license="Apache-2.0",
    author="Nanne Aben",
    author_email="nanne@kaiko.ai",
    description="Column-wise type annotations for pyspark DataFrames",
    keywords="pyspark spark typing type checking annotations",
    long_description=get_long_description(),
    long_description_content_type="text/x-rst",
    packages=find_packages(include=["typedspark", "typedspark.*"]),
    install_requires=get_requirements(),
    python_requires=">=3.9.0",
    classifiers=["Programming Language :: Python", "Typing :: Typed"],
    version_config=True,
    setup_requires=["setuptools-git-versioning"],
    package_data={"typedspark": ["py.typed"]},
    extras_require={
        "pyspark": ["pyspark"],
    },
)



"""Module containing classes and functions related to TypedSpark Columns."""

from logging import warn
from typing import Generic, Optional, Type, TypeVar, Union, get_args, get_origin

from pyspark.sql import Column as SparkColumn
from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.functions import col
from pyspark.sql.types import DataType

from typedspark._core.datatypes import StructType

T = TypeVar("T", bound=DataType)


class EmptyColumn(SparkColumn):
    """Column object to be instantiated when there is no active Spark session."""

    def __init__(self, *args, **kwargs) -> None:  # pragma: no cover
        pass


class Column(SparkColumn, Generic[T]):
    """Represents a ``Column`` in a ``Schema``. Can be used as:

    .. code-block:: python

        class A(Schema):
            a: Column[IntegerType]
            b: Column[StringType]
    """

    def __new__(
        cls,
        name: str,
        dataframe: Optional[DataFrame] = None,
        curid: Optional[int] = None,
        dtype: Optional[T] = None,
        parent: Union[DataFrame, "Column", None] = None,
    ):
        """``__new__()`` instantiates the object (prior to ``__init__()``).

        Here, we simply take the provided ``name``, create a pyspark
        ``Column`` object and cast it to a typedspark ``Column`` object.
        This allows us to bypass the pypsark ``Column`` constuctor in
        ``__init__()``, which requires parameters that may be difficult
        to access.
        """
        # pylint: disable=unused-argument

        if dataframe is not None and parent is None:
            parent = dataframe
            warn("The use of Column(dataframe=...) is deprecated, use Column(parent=...) instead.")

        column: SparkColumn
        if SparkSession.getActiveSession() is None:
            column = EmptyColumn()  # pragma: no cover
        elif parent is None:
            column = col(name)
        else:
            column = parent[name]

        column.__class__ = Column
        return column

    def __init__(
        self,
        name: str,
        dataframe: Optional[DataFrame] = None,
        curid: Optional[int] = None,
        dtype: Optional[T] = None,
        parent: Union[DataFrame, "Column", None] = None,
    ):
        # pylint: disable=unused-argument
        self.str = name
        self._dtype = dtype if dtype is not None else DataType
        self._curid = curid

    def __hash__(self) -> int:
        return hash((self.str, self._curid))

    @property
    def dtype(self) -> Type[T]:
        """Get the datatype of the column, e.g. Column[IntegerType] -> IntegerType."""
        dtype = self._dtype
        if get_origin(dtype) == StructType:
            dtype.schema = get_args(dtype)[0]  # type: ignore
            dtype.schema._parent = self  # type: ignore

        return dtype  # type: ignore
